import atexit
import inspect
import re
from types import UnionType
from typing import Optional, Union, get_args, get_origin, get_type_hints

from pgvector.psycopg import register_vector
import psycopg
from psycopg.rows import dict_row
from psycopg_pool import ConnectionPool

from idli import model_methods
from idli import sql_factory
from idli.errors import *
from idli.helpers import *
from idli.helpers import _BaseVector
from idli.internal import Column, Table


class Connection:

    def __init__(self, db_uri: str, sambar_dip: bool=False, extensions: list=[]):
        def configure(conn):
            if 'pgvector' in extensions:
                register_vector(conn)
            
        self._pool = ConnectionPool(db_uri, open=True, configure=configure)
        atexit.register(self._pool.close)
        
        self._sambar_dip = sambar_dip

        self.load_tables()
        self.load_columns()
        self.load_indexes()
        

    def exec_sql(self, *args):
        with self._pool.connection() as conn:
            return conn.execute(*args)
            
    
    def exec_sql_to_dict_rows(self, *args):
        with self._pool.connection() as conn:
            cur = conn.cursor(row_factory = dict_row)
            return cur.execute(*args)


    def load_tables(self):
        result = self.exec_sql_to_dict_rows(sql_factory.list_tables()).fetchall()
        self.__db_tables__ = { row['table_name']: Table(row['table_name']) for row in result }


    def load_columns(self):
        result = self.exec_sql_to_dict_rows(sql_factory.list_columns()).fetchall()
        for row in result:
            table_name = row['table_name']
            if table_name in self.__db_tables__:
                self.__db_tables__[table_name].add_column(Column.from_db_row(**row))


    def load_indexes(self):
        result = self.exec_sql_to_dict_rows(sql_factory.list_indexes()).fetchall()
        self.__db_indexes__ = { row['indexname']: row for row in result }
    

    def _ensure_table(self, cls):
        if cls.__table__.name not in self.__db_tables__:
            if self._sambar_dip:
                table_name = cls.__table__.name
                self.exec_sql(sql_factory.create_table(table_name))
                self.__db_tables__[table_name] = Table(table_name)
            else:
                raise TableNotFoundError(f'Table {cls.__tablename__} for model {cls.__name__} does not exist on database')


    def _build_column_model(self, cls):
        for key, val in get_type_hints(cls).items():
            if key.startswith('__'):
                continue
                
            col_name = key
            type_args = get_args(val)
            if type_args:
                if len(type_args)==1:
                    col_class = type_args[0]
                    nullable = False
                elif len(type_args)==2 and type_args[0] is type(None):
                    col_class = type_args[1]
                    nullable = True
                elif len(type_args)==2 and type_args[1] is type(None):
                    col_class = type_args[0]
                    nullable = True    
            else:
                col_class = val
                nullable = False
            
            default = getattr(cls, key, None)
            length = None
            if issubclass(col_class, _BaseVector):
                length = col_class.dimensions
                col_class = _BaseVector
                
            cls.__table__.add_column(Column.from_py_model(
                table_name = cls.__table__.name,
                name = col_name, 
                column_class = col_class, 
                nullable = nullable,
                default = default,
                length = length,
            ))


    def _reconcile_columns(self, cls):
        for column in cls.__table__.columns.values():
            db_table = self.__db_tables__[cls.__table__.name]
            db_column = db_table.columns.get(column.name)
            
            if db_column:
                if db_column.column_type != column.column_type:
                    raise ColumnTypeMismatchError(f"Column '{column.name}' is type '{db_column.column_type}' on database")
                
                if db_column.nullable == False and column.nullable == True:
                    if self._sambar_dip:
                        self.exec_sql(sql_factory.make_column_nullable(column))
                    else:
                        raise ColumnNotNullableError(f"Changing column '{column.name}' to nullable is not supported with sambar_dip=False")
                if db_column.nullable == True and column.nullable == False:
                    raise ColumnNullableError(f"Changing column '{column.name}' to not nullable is not supported")

                if db_column.default != column.default:
                    if self._sambar_dip:
                        self.exec_sql(sql_factory.set_default_column_value(column))
                    else:
                        raise ColumnDefaultMismatchError(f"Defined default value for column '{column.name}' does not match with the database")
            else:
                if self._sambar_dip:
                    self.exec_sql(sql_factory.create_column(column))
                else:
                    raise ColumnNotFoundError(f"Column '{column.name}' does not exist in table '{cls.__table__.name}'")


    def _handle_directives(self, cls):
        directives = getattr(cls, '__idli__', [])
        cls.__primary_key__ = ['id']
        cls.__indexes__ = {}

        for d in directives:
            if type(d) is PrimaryKey:
                cls.__primary_key__ = d.columns
            if type(d) is BTreeIndex:
                cls.__indexes__[f'{cls.__table__.name}_{d.name_hash}'] = d
            if type(d) is HNSWIndex:
                cls.__indexes__[f'{cls.__table__.name}_{d.name_hash}'] = d    


    def _reconcile_primary_key(self, cls):
        defined_pk_columns = cls.__primary_key__

        constraint = self.exec_sql_to_dict_rows(
            sql_factory.get_primary_key_constraint_name(cls.__table__.name),
        ).fetchall()

        if len(constraint) == 0:
            self.exec_sql(sql_factory.create_primary_key(
                table_name = cls.__table__.name,
                columns = defined_pk_columns,
            ))
            return

        constraint_name = constraint[0]['constraint_name']
            
        result = self.exec_sql_to_dict_rows(
            sql_factory.get_primary_key_columns(constraint_name)
        ).fetchall()
        existing_pk_columns = [c['column_name'] for c in result]

        reconciliation_required = False
        if len(defined_pk_columns) != len(existing_pk_columns):
            reconciliation_required = True
        else:
            for i in range(len(defined_pk_columns)):
                if defined_pk_columns[i] != existing_pk_columns[i]:
                    reconciliation_required = True
                    break
                    
        if reconciliation_required:    
            self.exec_sql(sql_factory.drop_constraint(
                table_name = cls.__table__.name,
                constraint_name = constraint_name,
            ))

            self.exec_sql(sql_factory.create_primary_key(
                table_name = cls.__table__.name,
                columns = defined_pk_columns,
            ))


    def _reconcile_indexes(self, cls):
        for name, idx in cls.__indexes__.items():
            if name in self.__db_indexes__:
                continue
            if type(idx) is BTreeIndex:
                self.exec_sql(sql_factory.create_btree_index(
                    table_name = cls.__table__.name,
                    columns = idx.columns,
                    index_name = name,
                ))
            if type(idx) is HNSWIndex:
                self.exec_sql(sql_factory.create_hnsw_index(
                    table_name = cls.__table__.name,
                    column = idx.column,
                    operation = idx.operation,
                    index_name = name,
                ))
            
    
    
    def Model(self, cls):
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', cls.__name__)
        s2 = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1)
        cls.__table__ = Table(s2.lower())

        self._ensure_table(cls)
        self._build_column_model(cls)
        self._reconcile_columns(cls)
        self._handle_directives(cls)
        self._reconcile_primary_key(cls)
        self._reconcile_indexes(cls)

        cls._connection = self
        cls.__init__ = model_methods.__init__
        cls._save_existing = model_methods._save_existing
        cls._save_new = model_methods._save_new
        cls.delete = model_methods.delete
        cls.save = model_methods.save
        cls.update = model_methods.update

        cls._obj_from_dict = classmethod(model_methods._obj_from_dict)
        cls.count = classmethod(model_methods.count)
        cls.select = classmethod(model_methods.select)
        
        return cls



from typing import List

from psycopg.sql import Identifier, Literal, SQL

from idli.helpers import *
from idli.internal import Column, Table


DATE_FMT = "%Y-%m-%d %H:%M:%S.%f"

OPERATORS = dict(
    eq = SQL('{} = {}'),
    gt = SQL('{} > {}'),
    gte = SQL('{} >= {}'),
    lt = SQL('{} < {}'),
    lte = SQL('{} <= {}'),
    neq = SQL('{} != {}'),
)


def create_btree_index(table_name: str, columns: List[str], index_name: str):
    return SQL(' ').join([
        SQL('CREATE INDEX IF NOT EXISTS {}').format(Identifier(index_name)),
        SQL('ON {} USING BTREE').format(Identifier(table_name)),
        SQL('').join([
            SQL('('),
            SQL(', ').join(
                [Identifier(c.strip('-')) + SQL(' ') + SQL('DESC' if c.startswith('-') else 'ASC') for c in columns]
            ),
            SQL(')'),
        ])
    ])


def create_hnsw_index(table_name: str, column: str, operation: str, index_name: str):
    op_code = {
        'l2d': 'vector_l2_ops',
        'l1d': 'vector_l1_ops',
        'inp': 'vector_ip_ops',
        'cos': 'vector_cosine_ops',
    }[operation]
    return SQL(' ').join([
        SQL('CREATE INDEX IF NOT EXISTS {}').format(Identifier(index_name)),
        SQL('ON {} USING HNSW').format(Identifier(table_name)),
        SQL('').join([
            SQL('('),
            Identifier(column),
            SQL(' '),
            SQL(op_code),
            SQL(')'),
        ])
    ])


def create_primary_key(table_name: str, columns: List[str]):
    return SQL(' ').join([
        SQL('ALTER TABLE {}').format(Identifier(table_name)),
        SQL('ADD CONSTRAINT {}').format(Identifier(table_name + '_pkey')),
        SQL('PRIMARY KEY'),
        SQL('').join([
            SQL('('),
            SQL(', ').join([Identifier(c) for c in columns]),
            SQL(')')
        ])
    ])


def create_column(column: Column):
    column_type = column.column_type
    if column_type == 'VECTOR' and column.length:
        column_type = f'VECTOR({column.length})'
    default = column.default
    if default == AutoInt and column_type == 'INTEGER':
        column_type = 'SERIAL'
        default = None
    elif default == AutoUUID and column_type == 'UUID':
        default = 'uuidv7()'
        
    stmt = [
        SQL('ALTER TABLE {}').format(Identifier(column.table_name)),
        SQL('ADD COLUMN IF NOT EXISTS {} {}').format(
            Identifier(column.name),
            SQL(column_type),
        ),
    ]
    
    if column.nullable == False:
        stmt.append(SQL('NOT NULL'))
        
    if default != None:
        if column.column_type == 'TIMESTAMP':
            stmt.append(SQL('DEFAULT {}').format(Literal(default.strftime(DATE_FMT))))
        elif default=='uuidv7()':
            stmt.append(SQL('DEFAULT uuidv7()'))
        else:
            stmt.append(SQL('DEFAULT {}').format(Literal(str(default))))

    return SQL(' ').join(stmt)


def create_table(table_name: str):
    return SQL('''
        CREATE TABLE IF NOT EXISTS {} ();
    ''').format(Identifier(table_name))


def count_by_filter(table_name: str, filters: dict):
    stmt = [SQL('SELECT COUNT(*) FROM {}').format(
        Identifier(table_name),
    )]

    # TODO DRY
    if (filters is not None) and len(filters):
        filter_bits = []
        for key, val in filters.items():
            if '__' in key:
                col, op = key.split('__')
                filter_bits.append(OPERATORS[op].format(Identifier(col), Literal(val)))
            else:
                if val is None:
                    filter_bits.append(SQL('{} IS NULL').format(Identifier(key)))
                else:
                    col, op = key, 'eq'
                    filter_bits.append(OPERATORS[op].format(Identifier(col), Literal(val)))
            
                
        stmt.append(SQL('WHERE ') + SQL(' AND ').join(filter_bits))
    
    return SQL(' ').join(stmt)


def delete_by_filter(table_name: str, filter: dict):
    # TODO DRY
    return SQL(' ').join([
        SQL('DELETE FROM {}').format(Identifier(table_name)),
        SQL(' ').join([
            SQL('WHERE'),
            SQL(', ').join([
                SQL('{} = {}').format(Identifier(c), Literal(v)) if v is not None else SQL('{} IS NULL').format(Identifier(c)) for c, v in filter.items()
            ]),
        ]),
    ])


def drop_constraint(table_name: str, constraint_name: str):
    return SQL('''
        ALTER TABLE {} DROP CONSTRAINT {};
    ''').format(Identifier(table_name), Identifier(constraint_name))


def get_primary_key_columns(constraint_name: str):
    return SQL('''
        SELECT column_name 
        FROM information_schema.key_column_usage 
        WHERE constraint_schema = 'public' AND constraint_name = {}
        ORDER BY ordinal_position;
    ''').format(Literal(constraint_name))


def get_primary_key_constraint_name(table_name: str):
    return SQL('''
        SELECT constraint_name
        FROM information_schema.table_constraints
        WHERE constraint_type = 'PRIMARY KEY' AND table_name = {};
    ''').format(Literal(table_name))


def insert_row(table_name: str, columns: List[str], values: List[str]):
    return SQL(' ').join([
        SQL('INSERT INTO {}').format(Identifier(table_name)),
        SQL('').join([
            SQL('('),
            SQL(', ').join([Identifier(c) for c in columns]),
            SQL(')'),
        ]),
        SQL('VALUES'),
        SQL('').join([
            SQL('('),
            SQL(', ').join([Literal(v) for v in values]),
            SQL(')'),
        ]),
    ])


def list_columns():
    return SQL("""
        SELECT 
            c.table_name, 
            c.column_name, 
            -- If it's a vector, name it 'vector'; otherwise, keep the standard name
            CASE 
                WHEN c.udt_name = 'vector' THEN 'vector' 
                ELSE c.data_type 
            END AS data_type,
            -- Extract dimensions safely (only if udt_name is vector)
            CASE 
                WHEN c.udt_name = 'vector' THEN NULLIF(a.atttypmod, -1)
                ELSE NULL 
            END AS vector_dimensions,
            c.is_nullable, 
            c.column_default
        FROM information_schema.columns c
        JOIN pg_class t ON c.table_name = t.relname
        JOIN pg_namespace n ON t.relnamespace = n.oid AND c.table_schema = n.nspname
        JOIN pg_attribute a ON t.oid = a.attrelid AND c.column_name = a.attname
        WHERE c.table_schema = 'public'
        ORDER BY c.table_name, c.ordinal_position;
    """)


def list_indexes():
    return SQL('''
        SELECT schemaname, tablename, indexname, indexdef
        FROM pg_indexes
        WHERE schemaname = 'public'
        ORDER BY tablename, indexname;
    ''')


def list_tables():
    return SQL("""
        SELECT table_name
        FROM information_schema.tables
        WHERE table_schema = 'public'
        AND table_type = 'BASE TABLE';
    """)
    

def make_column_nullable(column: Column):
    return SQL('ALTER TABLE {} ALTER COLUMN {} DROP NOT NULL').format(
        Identifier(column.table_name),
        Identifier(column.name),
    )


def query_rows(
        table_name: str,
        filters: dict | None = None,
        limit: int | None = None,
        skip: int | None = None,
        order_by: List | None = None,
    ):
    
    stmt = [SQL('SELECT * FROM {}').format(
        Identifier(table_name),
    )]

    if (filters is not None) and len(filters):
        filter_bits = []
        for key, val in filters.items():
            if '__' in key:
                col, op = key.split('__')
                filter_bits.append(OPERATORS[op].format(Identifier(col), Literal(val)))
            else:
                if val is None:
                    filter_bits.append(SQL('{} IS NULL').format(Identifier(key)))
                else:
                    col, op = key, 'eq'
                    filter_bits.append(OPERATORS[op].format(Identifier(col), Literal(val)))    
        stmt.append(SQL('WHERE ') + SQL(' AND ').join(filter_bits))
    
    if order_by is not None:
        ordering_bits = []
        for col in order_by:
            if type(col) is VNN:
                ordering_bits.append(SQL(' ').join([
                    Identifier(col.column),
                    SQL(col.operator),
                    Literal(str(list(col.vector))),
                ]))
            elif col.startswith('-'):
                ordering_bits.append(SQL('{} DESC').format(Identifier(col[1:])))
            else:
                ordering_bits.append(Identifier(col))
        stmt.append(SQL('ORDER BY ') + SQL(',').join(ordering_bits))

    if limit is not None:
        stmt.append(SQL('LIMIT {}').format(Literal(limit)))
    
    if skip is not None:
        stmt.append(SQL('OFFSET {}').format(Literal(skip)))
        
    return SQL(' ').join(stmt)


def set_default_column_value(column: Column):
    if column.default != None:
        if column.default == AutoUUID:
            return SQL('ALTER TABLE {} ALTER COLUMN {} SET DEFAULT uuidv7()').format(
                Identifier(column.table_name),
                Identifier(column.name),
            )
            
        if column.column_type == 'TIMESTAMP':
            default_str = column.default.strftime(DATE_FMT)
        else:
            default_str = str(column.default)
        
        return SQL('ALTER TABLE {} ALTER COLUMN {} SET DEFAULT {}').format(
            Identifier(column.table_name),
            Identifier(column.name),
            Literal(default_str),
        )
    else:
        return SQL('ALTER TABLE {} ALTER COLUMN {} DROP DEFAULT').format(
            Identifier(column.table_name),
            Identifier(column.name),
        )


def update_row(table_name: str, pk_filter: dict, updates: dict):
    return SQL(' ').join([
        SQL('UPDATE {}').format(Identifier(table_name)),
        SQL(' ').join([
            SQL('SET'),
            SQL(', ').join([SQL('{} = {}').format(Identifier(c), Literal(v)) for c, v in updates.items()]),
        ]),
        SQL(' ').join([
            SQL('WHERE'),
            SQL(', ').join([SQL('{} = {}').format(Identifier(c), Literal(v)) for c, v in pk_filter.items()]),
        ]),
    ])



import copy
from datetime import datetime
from typing import List, Sequence
from uuid import UUID

from idli import sql_factory
from idli.errors import InvalidColumnTypeError
from idli.helpers import *
from idli.helpers import _BaseVector


DATE_FMT = "%Y-%m-%d %H:%M:%S.%f"


class ColumnType:

    def __init__(self, py_type, db_type, py_to_db, db_to_py, db_val_to_py_val):
        self.py_type = py_type
        self.db_type = db_type
        self.py_to_db = py_to_db
        self.db_to_py = db_to_py
        self.db_val_to_py_val = db_val_to_py_val

    
COLUMN_TYPES = {
    'BOOLEAN': ColumnType(
        py_type = bool,
        db_type = 'boolean',
        py_to_db = lambda x: str(x) if x is not None else None,
        db_to_py = lambda x: x.lower()=='true',
        db_val_to_py_val = lambda x: x,
    ),
    'TIMESTAMP': ColumnType(
        py_type = datetime,
        db_type = 'timestamp without time zone',
        py_to_db = lambda x: x.strftime(DATE_FMT) if x is not None else None,
        db_to_py = lambda x: datetime.strptime(x, DATE_FMT),
        db_val_to_py_val = lambda x: x,
    ),
    'NUMERIC': ColumnType(
        py_type = float,
        db_type = 'numeric',
        py_to_db = lambda x: str(x) if x is not None else None,
        db_to_py = lambda x: float(x),
        db_val_to_py_val = lambda x: float(x) if x is not None else None,
    ),
    'INTEGER': ColumnType(
        py_type = int,
        db_type = 'integer',
        py_to_db = lambda x: str(x) if x is not None else None,
        db_to_py = lambda x: int(x),
        db_val_to_py_val = lambda x: x,
    ),
    'VARCHAR': ColumnType(
        py_type = str,
        db_type = 'character varying',
        py_to_db = lambda x: x if x is not None else None,
        db_to_py = lambda x: x,
        db_val_to_py_val = lambda x: x,
    ),
    'UUID': ColumnType(
        py_type = UUID,
        db_type = 'uuid',
        py_to_db = lambda x: str(x) if x is not None else None,
        db_to_py = lambda x: UUID(x),
        db_val_to_py_val = lambda x: x,
    ),
    'VECTOR': ColumnType(
        py_type = _BaseVector,
        db_type = 'vector',
        py_to_db = lambda x: list(x) if x is not None else None,
        db_to_py = lambda x: x,
        db_val_to_py_val = lambda x: x,
    ),
}

PY_COLUMN_TYPES = { COLUMN_TYPES[key].py_type: key for key in COLUMN_TYPES }
DB_COLUMN_TYPES = { COLUMN_TYPES[key].db_type: key for key in COLUMN_TYPES }



class Column:

    def __init__(
        self,
        table_name: str,
        name: str,
        column_type = None,
        nullable: bool = False,
        default = None,
        length = None, # Only for vector
    ):
        self.table_name = table_name
        self.name = name
        self.column_type = column_type
        self.nullable = nullable
        self.default = default
        self.length = length


    @staticmethod
    def from_py_model(
        table_name: str,
        name: str,
        column_class,
        nullable: bool = False,
        default = None,
        length = None,
    ):
        
        if column_class not in PY_COLUMN_TYPES:
            raise InvalidColumnTypeError(f"Unsupported class '{column_class.__name__}' for column '{name}'")
        
        return Column(
            table_name = table_name,
            name = name,
            column_type = PY_COLUMN_TYPES[column_class],
            nullable = nullable,
            default = default,
            length = length,
        )


    @staticmethod
    def from_db_row(
        table_name: str,
        column_name: str,
        data_type: str,
        is_nullable: str,
        column_default,
        vector_dimensions = None
    ):
            
        if data_type not in DB_COLUMN_TYPES:
            raise InvalidColumnTypeError(f"No handler for DB column {column_name} of type {date_type}")

        length = None
        if data_type == 'vector':
            length = vector_dimensions
        column_type = DB_COLUMN_TYPES[data_type]

        if column_default:
            if column_type=='BOOLEAN':
                column_default = True if column_default.lower()=='true' else False
            elif column_type=='INTEGER':
                default_for_auto = f"nextval('{table_name}_{column_name}_seq'::regclass)"
                if column_default == default_for_auto:
                    column_default = AutoInt
                else:
                    try:
                        column_default = int(column_default)
                    except:
                        pass
            elif column_type=='NUMERIC':
                try:
                    column_default = float(column_default)
                except:
                    pass
            elif column_type=="TIMESTAMP":
                try:
                    column_default = column_default.rsplit('::timestamp without time zone', 1)[0].strip("'")
                    if '.' not in column_default:
                        column_default += '.000000'
                    column_default = datetime.strptime(column_default, DATE_FMT)
                except Exception as e:
                    pass      
            elif column_type=="UUID":
                if column_default == 'uuidv7()':
                    column_default = AutoUUID
                else:
                    try:
                        column_default = UUID(column_default.rsplit('::uuid', 1)[0].strip("'"))
                    except Exception as e:
                        pass            
            elif column_type=='VARCHAR':
                column_default = column_default.rsplit('::character varying', 1)[0].strip("'")
            elif column_type=='VECTOR':
                column_default = list(map(
                    float,
                    column_default.rsplit('::vector', 1)[0].strip("'[]").split(',')
                ))

        
        return Column(
            table_name = table_name,
            name = column_name,
            column_type = column_type,
            nullable = is_nullable.lower()=='yes',
            default = column_default,
            length = length
        )
        

    def __repr__(self):
        return f'Column<{self.column_type}> {self.name}'


    def py_to_db(self, val):
        return COLUMN_TYPES[self.column_type].py_to_db(val)


    def db_val_to_py_val(self, db_val):
        return COLUMN_TYPES[self.column_type].db_val_to_py_val(db_val)



class QuerySet:

    def __init__(
        self, 
        cls, 
        filters: dict | None = None,
        limit: int | None = None,
        skip: int | None = None,
        order_by: List | None = None,
    ):
        self._cls = cls
        self._filters = filters
        self._cursor = None
        self._limit = limit
        self._skip = skip
        self._order_by = order_by


    def __iter__(self):
        self._cursor = self._cls._connection.exec_sql_to_dict_rows(
            sql_factory.query_rows(
                table_name = self._cls.__table__.name,
                filters = self._filters,
                limit = self._limit,
                skip = self._skip,
                order_by = self._order_by,
            )
        )
        for row in self._cursor:
            yield self._cls._obj_from_dict(row)
    

    def __getitem__(self, key):
        if isinstance(key, slice):
            new_qs = copy.copy(self)
            if (key.start is not None) and (key.stop is not None):
                new_qs._limit = key.stop - key.start
                new_qs._skip = key.start
            elif key.start is not None:
                new_qs._limit = None
                new_qs._skip = key.start
            elif key.stop is not None:
                new_qs._limit = key.stop
                new_qs._skip = None
            return new_qs


    def one(self):
        new_qs = copy.copy(self)
        new_qs._limit = 1
        for row in self:
            return row


    def order_by(self, *args):
        new_qs = copy.copy(self)
        if len(args)>0:
            new_qs._order_by = args
        return new_qs


class Table:

    def __init__(self, name: str):
        self.name = name
        self.columns = {}
    

    def __repr__(self):
        return f"Table {self.name}: {', '.join(self.columns.keys())}"
    

    def add_column(self, column: Column):
        self.columns[column.name] = column
    





from idli import sql_factory
from idli.errors import InvalidValueTypeError
from idli.helpers import AutoInt, AutoUUID, _BaseVector
from idli.internal import PY_COLUMN_TYPES, QuerySet


def count(cls, **kwargs):
    return list(cls._connection.exec_sql_to_dict_rows(sql_factory.count_by_filter(
        table_name = cls.__table__.name,
        filters = kwargs,
    )))[0]['count']


def select(cls, **kwargs):
    return QuerySet(cls, filters=kwargs)


def _obj_from_dict(cls, row_dict):
    obj = cls()
    obj.__original__ = {}
    for column_name in cls.__table__.columns:
        column = cls.__table__.columns[column_name]
        value = column.db_val_to_py_val(
            row_dict.get(column_name)
        )
        setattr(obj, column_name, value)
        obj.__original__[column_name] = value
    return obj


def __init__(self, **kwargs):
    self.__original__ = {}
    for key in kwargs:
        if key in self.__table__.columns:
            setattr(self, key, kwargs[key])


def _save_existing(self):
    updates = {}
    pk_filter = {}
    for key in self.__table__.columns:
        column = self.__table__.columns[key]
        if hasattr(self, key):
            val = getattr(self, key)
            val_type = type(val)
            if column.column_type == 'VECTOR':
                val_type = _BaseVector
                
            if val is None:
                if column.nullable:
                    pass
                else:
                    raise CannotBeNoneError(f"Value for column '{key}' cannot be None")
            else:
                if val_type not in PY_COLUMN_TYPES:
                    raise InvalidValueTypeError(f"Invalid value '{val}' for column '{key}'")
                elif not (column.column_type == PY_COLUMN_TYPES[val_type] or val is None):
                    raise InvalidValueTypeError(f"Invalid value '{val}' for column '{key}'")

            if key in self.__class__.__primary_key__:
                pk_filter[key] = column.py_to_db(val)
            else:
                updates[key] = column.py_to_db(val)
                
    self._connection.exec_sql(sql_factory.update_row(
        table_name = self.__table__.name,
        pk_filter = pk_filter,
        updates = updates,
    ))


def _save_new(self):
    columns = []
    values = []
    for key in self.__table__.columns:
        column = self.__table__.columns[key]
        if hasattr(self, key):
            val = getattr(self, key)
            if val not in [AutoInt, AutoUUID, None]:
                val_type = type(val)
                if column.column_type == 'VECTOR':
                    val_type = _BaseVector
                
                if val_type not in PY_COLUMN_TYPES:
                    raise InvalidValueTypeError(f"Invalid value '{val}' for column '{key}'")
                if column.column_type != PY_COLUMN_TYPES[val_type]:
                    raise InvalidValueTypeError(f"Invalid value '{val}' for column '{key}'")
                columns.append(key)
                values.append(column.py_to_db(val))
                
    self._connection.exec_sql(sql_factory.insert_row(
        table_name = self.__table__.name,
        columns = columns,
        values = values,
    ))

                

def delete(self):
    pk = {}
    for key in self.__class__.__primary_key__:
        column = self.__table__.columns[key]
        pk[key] = column.py_to_db(getattr(self, key))
        
    self._connection.exec_sql(sql_factory.delete_by_filter(
        table_name = self.__table__.name,
        filter = pk,
    ))


def save(self):
    if len(self.__original__.keys()) == 0:
        self._save_new()
    else:
        self._save_existing()


def update(self, **kwargs):
    for key in kwargs:
        setattr(self, key, kwargs[key])
    self.save()




from typing import Annotated, Sequence

class AutoInt:
    pass


class AutoUUID:
    pass


class BTreeIndex:

    def __init__(self, *args):
        self.columns = list(args)


    @property
    def name_hash(self):
        return '_'.join(map(
            lambda x: 'd' + x[1:] if x.startswith('-') else 'a' + x,
            self.columns
        )) + '_btree'


class HNSWIndex:

    def __init__(self, column: str, operation: str):
        self.column = column
        self.operation = operation


    @property
    def name_hash(self):
        return f'{self.column}_{self.operation}_hnsw'


class PrimaryKey:

    def  __init__(self, *args):
        self.columns = list(args)


class _BaseVector(list):
    """Base class for all vector types in the ORM."""
    dimensions: int = 0

    def __init__(self, iterable=None):
        if iterable is not None:
            super().__init__(iterable)
            if self.dimensions > 0 and len(self) != self.dimensions:
                raise ValueError(
                    f"Expected {self.dimensions} dimensions, got {len(self)}"
                )

    def __repr__(self):
        content = super().__repr__()
        return f"Vector({self.dimensions})({content})"


class VNN:

    def __init__(self, column, operator, vector):
        self.column = column
        self.operator = operator
        self.vector = vector


    @classmethod
    def l2d(cls, column, vector):
        return cls(column=column, operator='<->', vector=vector)


    @classmethod
    def inp(cls, column, vector):
        return cls(column=column, operator='<#>', vector=vector)


    @classmethod
    def cos(cls, column, vector):
        return cls(column=column, operator='<=>', vector=vector)


    @classmethod
    def l1d(cls, column, vector):
        return cls(column=column, operator='<+>', vector=vector)
    

def Vector(dimensions: int):
    return type(f"Vector_{dimensions}", (_BaseVector,), {"dimensions": dimensions})




class CannotBeNoneError(Exception):
    pass
    
class ColumnDefaultMismatchError(Exception):
    pass

class ColumnNotFoundError(Exception):
    pass

class ColumnNullableError(Exception):
    pass

class ColumnNullableError(Exception):
    pass

class ColumnTypeMismatchError(Exception):
    pass

class InvalidColumnTypeError(Exception):
    pass

class InvalidValueTypeError(Exception):
    pass

class TableNotFoundError(Exception):
    pass

